/*	
 *	clipmp3.c
 *
 *	P/ECE MP3 Driver
 *
 *	CLiP - Common Library for P/ECE
 *	Copyright (C) 2001-2003 Naoyuki Sawa
 *
 *	* Tue Oct 28 08:11:00 JST 2003 Naoyuki Sawa
 *	- 1st [XB
 */
#include "clip.h"
#include "clipmp3.h" /* TODO:Cuɑgݍ񂾂O܂ */

/****************************************************************************
 *	t[wb_
 ****************************************************************************/

/* t[wb_ǂݍ݂܂B
 * [in]
 *	bs		rbgXg[\́B
 *	frame		t[wb_i[\́B
 */
void
mp3_read_frame_header(MP3BITSTREAM* bs, MP3FRAMEHEADER* frame)
{
	frame->syncword			= mp3_bitstream_get(bs, 12);
	frame->id			= mp3_bitstream_get(bs,  1);
	frame->layer			= mp3_bitstream_get(bs,  2);
	frame->protection_bit		= mp3_bitstream_get(bs,  1);
	frame->bitrate_index		= mp3_bitstream_get(bs,  4);
	frame->sampling_frequency	= mp3_bitstream_get(bs,  2);
	frame->padding_bit		= mp3_bitstream_get(bs,  1);
	frame->private_bit		= mp3_bitstream_get(bs,  1);
	frame->mode			= mp3_bitstream_get(bs,  2);
	frame->mode_extention		= mp3_bitstream_get(bs,  2);
	frame->copyright		= mp3_bitstream_get(bs,  1);
	frame->original			= mp3_bitstream_get(bs,  1);
	frame->emphasis			= mp3_bitstream_get(bs,  2);

	ASSERT(frame->syncword == 0xfff);	/* [h */
	ASSERT(frame->id == 1);			/* MPEG1 */
	ASSERT(frame->layer == 1);		/* LayerIII */
	ASSERT(frame->protection_bit == 1);	/* CRCیȂ */
	ASSERT(frame->padding_bit == 0);	/* pfBOrbgȂ */
}

/****************************************************************************
 *	Oj[
 ****************************************************************************/

/* Oj[ǂݍ݂܂B
 * [in]
 *	bs		rbgXg[\́B
 *	granule		Oj[i[\́B
 */
void
mp3_read_granule_info(MP3BITSTREAM* bs, MP3GRANULEINFO* granule)
{
	granule->part2_3_length		= mp3_bitstream_get(bs, 12);
	granule->big_values		= mp3_bitstream_get(bs,  9);
	granule->global_gain		= mp3_bitstream_get(bs,  8);
					  mp3_bitstream_skip(bs,  4);
					  mp3_bitstream_skip(bs,  1);
					  mp3_bitstream_skip(bs, 22);
					  mp3_bitstream_skip(bs,  1);
					  mp3_bitstream_skip(bs,  1);
					  mp3_bitstream_skip(bs,  1);
}

/****************************************************************************
 *	TCh
 ****************************************************************************/

/* TChǂݍ݂܂B
 * [in]
 *	bs		rbgXg[\́B
 *	side		TChi[\́B
 *	channels	`lB
 */
void
mp3_read_side_info(MP3BITSTREAM* bs, MP3SIDEINFO* side, int channels)
{
	int i_granule;
	int i_channel;

	side->main_data_begin		= mp3_bitstream_get(bs, 9);
	switch(channels) {
	case 1:
		side->private_bits	= mp3_bitstream_get(bs, 5);
		side->scfsi		= mp3_bitstream_get(bs, 4);
		break;
	case 2:
		side->private_bits	= mp3_bitstream_get(bs, 3);
		side->scfsi		= mp3_bitstream_get(bs, 8);
		break;
	}
	for(i_granule = 0; i_granule < 2; i_granule++) {
		for(i_channel = 0; i_channel < channels; i_channel++) {
			mp3_read_granule_info(bs, &side->granule_info[i_granule][i_channel]);
		}
	}
}

/****************************************************************************
 *	MP3hCo
 ****************************************************************************/

#define MP3BUFLEN	(64 * 5)

int
mp3_init(MP3DRIVER* mp3, const void* data, int len)
{
	/* ܂[NAB */
	memset(mp3, 0, sizeof(MP3DRIVER));

	/* ǂݍ݈ʒuƏI[ʒui[܂B */
	mp3->pos = data;
	mp3->end = mp3->pos + len;

	/* ŏ̃t[fR[hAobt@𖞂܂B */
	if(mp3_decode_frame(mp3) != 0) return -1;

	return 0;
}

int
mp3_stream_callback(short* wbuff, int param)
{
	MP3DRIVER* mp3 = (MP3DRIVER*)param;
	short* wbuff_end = wbuff + MP3BUFLEN;
	int sampling_frequency = mp3->sampling_frequency;
	int i_fbuff = mp3->i_fbuff;
	int d_fbuff = mp3->d_fbuff;
	short* fbuff = mp3->fbuff;

	while(wbuff < wbuff_end) {
		*wbuff++ = fbuff[i_fbuff];
		d_fbuff += sampling_frequency;
		while(d_fbuff >= 16000) {
			d_fbuff -= 16000;
			i_fbuff++;
		}
		while(i_fbuff >= MP3FRAME_SIZE_HALF) {
			if(mp3_decode_frame(mp3) != 0) return -1;
			i_fbuff -= MP3FRAME_SIZE_HALF;
			sampling_frequency = mp3->sampling_frequency;
		}
	}

	mp3->i_fbuff = i_fbuff;
	mp3->d_fbuff = d_fbuff;

	return 0;
}

/* 1t[fR[h܂B
 * [in]
 *	mp3			MP3hCo\́B
 * [out]
 *	߂l			Ȃ0Ԃ܂B
 *				sȂ0ȊO̒lԂ܂B(I[ɒBꍇȂ)
 *	mp3->fbuff		fR[ht[f[^i[܂B
 *	mp3->sampling_frequency	fR[ht[̃TvOgi[܂B
 */
int
mp3_decode_frame(MP3DRIVER* mp3)
{
	MP3BITSTREAM bs, bs_tmp;
	MP3FRAMEHEADER frame;
	MP3SIDEINFO side;
	MP3GRANULEINFO* granule;
	int bitrate, sampling_frequency, channels, frame_size;
	int i_granule, i_channel;
	short* out;

	/* t[ԂŕۑKv̂Ȃobt@B */
#if 0	/* SRAMɎꍇ */
	static short buf1[GRANULE_SIZE];
	static short buf2[GRANULE_SIZE];
#else	/* RAMɎꍇ */
	static short* buf1 = (short*)0x2000 - GRANULE_SIZE * 1;
	static short* buf2 = (short*)0x2000 - GRANULE_SIZE * 2;
	extern unsigned long __END_FRAM_ADDR[];
	if((int)buf2 < (int)__END_FRAM_ADDR) DIE();
#endif

	/* f[^I[? */
	if(mp3->pos >= mp3->end) return -1;

	/* rbgXg[J܂B */
	mp3_bitstream_open(&bs, mp3->pos);

	/* t[wb_ǂݍ݂܂B */
	mp3_read_frame_header(&bs, &frame);
	bitrate = bitrate_table[frame.bitrate_index];
	sampling_frequency = sampling_frequency_table[frame.sampling_frequency];
	channels = frame.mode == 3 ? 1 : 2;

	/* TChǂݍ݂܂B */
	mp3_read_side_info(&bs, &side, channels);

	/* eOj[fR[h܂B */
	out = mp3->fbuff;
	for(i_granule = 0; i_granule < GRANULE_COUNT; i_granule++) {
		for(i_channel = 0; i_channel < channels; i_channel++) {
			granule = &side.granule_info[i_granule][i_channel];
			if(i_channel == 0) { /* `l̂ݎgp */
				bs_tmp = bs;
				/* nt}B */
				mp3_huffman_decode(&bs_tmp, granule, buf1);
				/* tʎqB */
				mp3_requantize(granule, buf1);
				/* GAVO팸B */
				//alias_reduce(buf1);ȗ
				/* IMDCTB */
				mp3_imdct(buf1, buf2, &mp3->imdct_save[0]);
				/* TuohB */
				mp3->subband_synthesys_offset = mp3_subband_synthesys(buf2, out, &mp3->subband_synthesys_save[0][0], mp3->subband_synthesys_offset);
				out += GRANULE_SIZE_HALF;
			}
			mp3_bitstream_skip(&bs, granule->part2_3_length);
		}
	}

	/* TvOgi[܂B */
	mp3->sampling_frequency = sampling_frequency / 2/*=HALF*/;

	/* t[ʒui߂܂B */
	frame_size = 144 * bitrate / sampling_frequency;
	mp3->pos += frame_size;

	return 0;
}

/****************************************************************************
 *	AvP[Vp֐
 ****************************************************************************/

int
mp3_play(const void* data, int len)
{
	static MP3DRIVER mp3; /* STATICł! */

	/* ܂mɒ~܂B */
	mp3_stop();

	/* MP3hCo܂B */
	if(mp3_init(&mp3, data, len) != 0) return -1;

	/* Xg[ĐJn܂B */
	stream_play(MP3BUFLEN, mp3_stream_callback, (int)&mp3);

	return 0;
}

void
mp3_stop()
{
	/* Xg[Đ~܂B */
	stream_stop();

	/* MP3hCõN[Abv͕svłB */
}

