/*	
 *	clipmp3.h
 *
 *	P/ECE MP3 Driver
 *
 *	CLiP - Common Library for P/ECE
 *	Copyright (C) 2001-2003 Naoyuki Sawa
 *
 *	* Tue Oct 28 08:11:00 JST 2003 Naoyuki Sawa
 *	- 1st [XB
 */
#ifndef __CLIP_MP3_H__
#define __CLIP_MP3_H__

/****************************************************************************
 *	萔
 ****************************************************************************/

#define SUBBAND_SIZE		18					/* 1Tuoh̃Tv */
#define SUBBAND_COUNT		32					/* 1t[̃Tuoh */
#define GRANULE_SIZE		(SUBBAND_SIZE * SUBBAND_COUNT)		/* 1Oj[̃Tv */
#define GRANULE_COUNT		2					/* 1t[̃Oj[ */
#define MP3FRAME_SIZE		(GRANULE_SIZE * GRANULE_COUNT)		/* 1t[̃Tv */

#define SUBBAND_SIZE_2		(SUBBAND_SIZE * 2)			/* IMDCTŎgp */
#define SUBBAND_COUNT_2		(SUBBAND_COUNT * 2)			/* TuohŎgp */

#define PRE_REQUANTIZE_MAX	(((1 << 4) - 1) + ((1 << 13) - 1))	/* nt}o͂̍ől */

/* * P/ECẼXs[J[o͂16KHzAMP3̃TvOg͍Œł32KHzB
 *   ]ĖɍĐ[g𔼕ɂł܂B
 *   2003/10/28݂̃vÓAfR[hŏIiK̃Tuoh(ԏd!)̂ƂŁA
 *   fR[h悤ɂďx҂ł܂B
 *   ƑO̒iKŔɂł΂̂łA@킩܂łcB
 * * ȉ̃V{́AĐ[g𔼕ɂ邱Ƃŉe󂯂ƂŎgĂ܂B
 */
#define SUBBAND_COUNT_HALF	(SUBBAND_COUNT / 2)
#define SUBBAND_COUNT_2_HALF	(SUBBAND_COUNT_2 / 2)
#define GRANULE_SIZE_HALF	(GRANULE_SIZE / 2)
#define MP3FRAME_SIZE_HALF	(MP3FRAME_SIZE / 2)

/****************************************************************************
 *	rbgXg[
 ****************************************************************************/

typedef struct _MP3BITSTREAM {
	int cnt;			/* buf̗Lrbg */
	unsigned int buf;		/* bit0`(cnt-1)܂ŗL */
	const unsigned char* pos;	/* ̓ǂݍ݃AhX */
} MP3BITSTREAM;

void mp3_bitstream_open(MP3BITSTREAM* bs, const void* pos);
int mp3_bitstream_get(MP3BITSTREAM* bs, int cnt);
void mp3_bitstream_skip(MP3BITSTREAM* bs, int cnt);

/****************************************************************************
 *	t[wb_
 ****************************************************************************/

typedef struct _MP3FRAMEHEADER {
	int syncword;			/* 12 */
	int id;				/*  1 */
	int layer;			/*  2 */
	int protection_bit;		/*  1 */
	int bitrate_index;		/*  4 */
	int sampling_frequency;		/*  2 */
	int padding_bit;		/*  1 */
	int private_bit;		/*  1 */
	int mode;			/*  2 */
	int mode_extention;		/*  2 */
	int copyright;			/*  1 */
	int original;			/*  1 */
	int emphasis;			/*  2 */
} MP3FRAMEHEADER;			/*=32 */

extern const int bitrate_table[16/*bitrate_index*/];
extern const int sampling_frequency_table[4/*sampling_frequency*/];

void mp3_read_frame_header(MP3BITSTREAM* bs, MP3FRAMEHEADER* frame);

/****************************************************************************
 *	Oj[
 ****************************************************************************/

typedef struct _MP3GRANULEINFO {
	int part2_3_length;			/* 12 */
	int big_values;				/*  9 */
	int global_gain;			/*  8 */
	int scalefac_compress;			/*  4 */
	int windows_witching_flag;		/*  1 */
	/* switch(windows_witching_flag)	  {0:  | 1: }*/
	int block_type;				/*     | 2   */
	int mixed_block_flag;			/*     | 1   */
	int table_select[3];			/* 5x3 | 5x2 */
	int subblock_gain[3];			/*     | 3x3 */
	int region0_count;			/* 4   |     */
	int region1_count;			/* 3   |     */
	int preflag;				/*  1 */
	int scalefac_scale;			/*  1 */
	int count1table_select;			/*  1 */
} MP3GRANULEINFO;				/*=59 */

void mp3_read_granule_info(MP3BITSTREAM* bs, MP3GRANULEINFO* granule);

/****************************************************************************
 *	TCh
 ****************************************************************************/

typedef struct _MP3SIDEINFO {
	int main_data_begin;			/* 9 */
	/* switch(channels)			  {1:     | 2:    }*/
	int private_bits;			/*  5     |  3     */
	int scfsi;				/*  4     |  8     */
	MP3GRANULEINFO granule_info[2][2];	/* 59x2x1 | 59x2x2 */
} MP3SIDEINFO;					/*=136      256    */

void mp3_read_side_info(MP3BITSTREAM* bs, MP3SIDEINFO* side, int channels);

/****************************************************************************
 *	nt}
 ****************************************************************************/

extern const short huffman_table16[][2];

void mp3_huffman_decode(MP3BITSTREAM* bs, const MP3GRANULEINFO* granule, short* out/*[SUBBAND_COUNT][SUBBAND_SIZE]*/);

/****************************************************************************
 *	tʎq
 ****************************************************************************/

extern const int gain_table[256/*GRANULE.global_gain*/];
extern const short pow1_3_table[PRE_REQUANTIZE_MAX + 1];

void mp3_requantize(MP3GRANULEINFO* granule/*[SUBBAND_COUNT][SUBBAND_SIZE]*/, short* in_out);

/****************************************************************************
 *	IMDCT
 ****************************************************************************/

extern const short imdct_matrix[SUBBAND_SIZE_2][SUBBAND_SIZE];

void mp3_imdct(const short* in/*[SUBBAND_COUNT][SUBBAND_SIZE]*/, short* out/*[SUBBAND_SIZE][SUBBAND_COUNT]*/, short* save/*[SUBBAND_COUNT][SUBBAND_SIZE]*/);

/****************************************************************************
 *	Tuoh
 ****************************************************************************/

extern const short subband_matrix[SUBBAND_COUNT_2_HALF][SUBBAND_COUNT];
extern const short polyphase_filter[SUBBAND_COUNT_HALF][16];

int mp3_subband_synthesys(const short* in/*[SUBBAND_SIZE][SUBBAND_COUNT]*/, short* out/*[SUBBAND_SIZE][SUBBAND_COUNT_HALF]*/, short* save/*[16][SUBBAND_COUNT_2_HALF]*/, int offset);

/****************************************************************************
 *	MP3hCo
 ****************************************************************************/

typedef struct _MP3DRIVER {
	const unsigned char* pos;
	const unsigned char* end;
	/* t[ԂŕۑKv̂obt@ */
	short imdct_save[GRANULE_SIZE];
	short subband_synthesys_save[16][SUBBAND_COUNT_2_HALF];
	int subband_synthesys_offset;
	/* t[[gϊpobt@ */
	int sampling_frequency;
	int i_fbuff;
	int d_fbuff;
	short fbuff[MP3FRAME_SIZE_HALF];
} MP3DRIVER;

int mp3_init(MP3DRIVER* mp3, const void* data, int len);
int mp3_stream_callback(short* wbuff, int param);
int mp3_decode_frame(MP3DRIVER* mp3);

/****************************************************************************
 *	AvP[Vp֐
 ****************************************************************************/

int mp3_play(const void* data, int len);
void mp3_stop();

#endif /*__CLIP_MP3_H__*/
