/*	
 *	frammp3.c
 *
 *	P/ECE MP3 Driver
 *
 *	CLiP - Common Library for P/ECE
 *	Copyright (C) 2001-2003 Naoyuki Sawa
 *
 *	* Tue Oct 28 08:11:00 JST 2003 Naoyuki Sawa
 *	- 1st [XB
 */
#include "clip.h"
#include "clipmp3.h" /* TODO:Cuɑgݍ񂾂O܂ */

/****************************************************************************
 *	rbgXg[
 ****************************************************************************/

/* rbgXg[J܂B
 * [in]
 *	bs		rbgXg[\́B
 *	pos		ŏ̓ǂݍ݃AhXB
 */
void
mp3_bitstream_open(MP3BITSTREAM* bs, const void* pos)
{
	bs->cnt = 0;
	/*bs->buf = 0;{{sv}}*/
	bs->pos = (const unsigned char*)pos;
}

/* rbgXg[rbgǂݍ݂܂B
 * [in]
 *	bs		rbgXg[\́B
 *	cnt		ǂݍރrbgB(0`32)
 * [out]
 *	߂l		ǂݍ񂾃rbgB
 */
int
mp3_bitstream_get(MP3BITSTREAM* bs, int cnt)
{
	int bits;

	ASSERT(0 <= cnt && cnt <= 32);

	if(cnt <= 24) {
		while(bs->cnt < cnt) {
			bs->buf = bs->buf << 8 | *bs->pos++;
			bs->cnt += 8;
		}
		bs->cnt -= cnt;
		bits = (bs->buf >> bs->cnt) & ((1 << cnt) - 1);
	} else {
		bits = mp3_bitstream_get(bs,       24) << 8 |
		       mp3_bitstream_get(bs, cnt - 24);
	}

	return bits;
}

/* rbgXg[rbgǂݔ΂܂B
 * [in]
 *	bs		rbgXg[\́B
 *	cnt		ǂݔ΂rbgB(0`C)
 */
void
mp3_bitstream_skip(MP3BITSTREAM* bs, int cnt)
{
	if(bs->cnt >= cnt) {
		bs->cnt -= cnt;
	} else {
		cnt -= bs->cnt;
		bs->pos += cnt / 8;
		cnt &= 7;
		if(cnt) {
			bs->buf = *bs->pos++;
			bs->cnt = 8 - cnt;
		} else {
			bs->cnt = 0;
		}
	}
}

/****************************************************************************
 *	nt}
 ****************************************************************************/

/* nt}B
 * [in]
 *	bs		rbgXg[\́B
 *	granule		Oj[\́B
 *	out		o̓obt@B
 */
void
mp3_huffman_decode(MP3BITSTREAM* bs, const MP3GRANULEINFO* granule, short* out/*[SUBBAND_COUNT][SUBBAND_SIZE]*/)
{
	int i, n, x, y, linbits, i_node;
	const short (*table)[2];

	table = huffman_table16;/*TODO:*/
	linbits = 13;		/*TODO:*/

	i = 0;

	/* BigValuë */
	n = granule->big_values * 2;
	while(i < n) {
		i_node = 0;
		do {
			i_node = table[i_node][mp3_bitstream_get(bs, 1)];
		} while(i_node > 0);
		i_node = -i_node;
		x = i_node >> 4;
		y = i_node & 15;
		if(x == 15) x += mp3_bitstream_get(bs, linbits);
		if(x && mp3_bitstream_get(bs, 1)) x = -x;
		if(y == 15) y += mp3_bitstream_get(bs, linbits);
		if(y && mp3_bitstream_get(bs, 1)) y = -y;
		*out++ = x;
		*out++ = y;
		i += 2;
	}

	/* Count1̈ */
	/* TODO: */

	/* rzerö */
	while(i < GRANULE_SIZE) {
		*out++ = 0;
		i++;
	}
}

/****************************************************************************
 *	tʎq
 ****************************************************************************/

/* tʎqB
 * [in]
 *	granule		Oj[\́B
 *	in_out		o̓obt@B
 */
void
mp3_requantize(MP3GRANULEINFO* granule/*[SUBBAND_COUNT][SUBBAND_SIZE]*/, short* in_out)
{
	int gain, i, w;
	short h;

	gain = gain_table[granule->global_gain];	/* 1.13.18 */
	for(i = 0; i < GRANULE_SIZE; i++) {
		h = *in_out;				/* 1.15. 0 */
		w = h * pow1_3_table[abs(h)];		/* 1.21.10 (short~short) */
		w = w * gain;				/* 1. 3.28 (int~int) */
		*in_out++ = w >> 14;			/* 1. 1.14 (I[o[t[) */
	}
}

/****************************************************************************
 *	IMDCT
 ****************************************************************************/

/* IMDCTB
 * [in]
 *	in		̓obt@B
 *	out		o̓obt@B
 *	save		t[ԂŕۑKv̂obt@B
 * [note]
 *	* Tuoḧꕔ肵Ă܂B
 */
void
mp3_imdct(const short* in/*[SUBBAND_COUNT][SUBBAND_SIZE]*/, short* out/*[SUBBAND_SIZE][SUBBAND_COUNT]*/, short* save/*[SUBBAND_COUNT][SUBBAND_SIZE]*/)
{
#if 0	/********************************************************************/

	int i, j, k, sum;
	const short* m;

	for(k = 0; k < SUBBAND_COUNT; k++) {
		m = &imdct_matrix[0][0];

		/* uÕOj[̌㔼vƁũ݂Oj[̑OvďóB */
		for(i = 0; i < SUBBAND_SIZE; i++) {
			sum = 0;
			for(j = 0; j < SUBBAND_SIZE; j++) {
				sum += *in++ * *m++;	/* short~short 1.3.28 */
			}
			sum >>= 14;	/* 1.1.14 */
			sum += *save++;	/* 1.1.14 */
			/* TuoḧꕔBs𕄍]ē]uĂB */
			if(k & i & 1) sum = -sum;
			*out = sum;	/* 1.1.14 */
			out += SUBBAND_COUNT;
			in -= SUBBAND_SIZE;
		}

		out -= GRANULE_SIZE - 1;
		save -= SUBBAND_SIZE;

		/* ũOj[̑Ovɍ邽߂Ɂũ݂Oj[̌㔼vۑB */
		for(i = 0; i < SUBBAND_SIZE; i++) {
			sum = 0;
			for(j = 0; j < SUBBAND_SIZE; j++) {
				sum += *in++ * *m++;	/* short~short 1.3.28 */
			}
			sum >>= 14;	/* 1.1.14 */
			*save++ = sum;	/* 1.1.14 */
			in -= SUBBAND_SIZE;
		}

		in += SUBBAND_SIZE;
	}

#else	/********************************************************************/

	asm volatile ("
	xld.w %r7, 32			; k = SUBBAND_COUNT
mp3_imdct_L10:
	xld.w %r11, imdct_matrix	; m = imdct_matrix
	;
	; uÕOj[̌㔼vƁũ݂Oj[̑OvďóB
	;
	xld.w %r6, 18			; i = SUBBAND_SIZE
mp3_imdct_L20:
	ld.w %alr, %r8			; ݍ
	xld.w %r10, 18			; SUBBAND_SIZE
	mac %r10	
	ld.w %r10, %alr			; sum >>= 14
	xsra %r10, 14
	ld.h %r15, [%r14]+		; sum += *save++
	add %r10, %r15
	;
	xld.w %r15, 1			; if(k & i & 1) sum = -sum
	and %r15, %r7
	and %r15, %r6
	xjreq mp3_imdct_L30
	not %r10, %r10
	xadd %r10, %r10, 1
mp3_imdct_L30:
	ld.h [%r13], %r10		; *out = sum
	xadd %r13, %r13, 64		; out += SUBBAND_COUNT * sizeof(short)
	xsub %r12, %r12, 36		; in -= SUBBAND_SIZE * sizeof(short)
	;
	xsub %r6, %r6, 1
	xjrne mp3_imdct_L20
	;
	xsub %r13, %r13, 1150		; out -= (GRANULE_SIZE - 1) * sizeof(short)
	xsub %r14, %r14, 36		; save -= SUBBAND_SIZE * sizeof(short)
	;
	; ũOj[̑Ovɍ邽߂Ɂũ݂Oj[̌㔼vۑB
	;
	xld.w %r6, 18			; i = SUBBAND_SIZE
mp3_imdct_L40:
	ld.w %alr, %r8			; ݍ
	xld.w %r10, 18			; SUBBAND_SIZE
	mac %r10	
	ld.w %r10, %alr			; sum >>= 14
	xsra %r10, 14
	ld.h [%r14]+, %r10		; *save++ = sum
	xsub %r12, %r12, 36		; in -= SUBBAND_SIZE * sizeof(short)
	;
	xsub %r6, %r6, 1
	xjrne mp3_imdct_L40
	;
	xadd %r12, %r12, 36		; in += SUBBAND_SIZE * sizeof(short)
	;
	xsub %r7, %r7, 1
	xjrne mp3_imdct_L10
	");

#endif	/********************************************************************/
}

/****************************************************************************
 *	Tuoh
 ****************************************************************************/

/* TuohB
 * [in]
 *	in		̓obt@B
 *	out		o̓obt@B
 *	save		t[ԂŕۑKv̂obt@B
 *	offset		saveItZbgB
 *			Omp3_subband_synthesys()̖߂lnĂB
 * [out]
 *	߂l		saveItZbgB
 * [note]
 *	* Tuoḧꕔmp3_imdct()ŐɍsĂ܂B
 */
int
mp3_subband_synthesys(const short* in/*[SUBBAND_SIZE][SUBBAND_COUNT]*/, short* out/*[SUBBAND_SIZE][SUBBAND_COUNT_HALF]*/, short* save/*[16][SUBBAND_COUNT_2_HALF]*/, int offset)
{
#if 0	/********************************************************************/

	int i, j, k, sum;
	const short* m;

#define OFFSET_MASK	(SUBBAND_COUNT_2_HALF * 16 - 1)

	for(k = 0; k < SUBBAND_SIZE; k++) {
		/* ݍ݉ZB(32f[^ԗ̈ɕϊ)
		 * ͍s̊s𕄍]ē]u鏈́AIMDCT̒ōsĂ܂B
		 */
		m = &subband_matrix[0][0];
		for(i = 0; i < SUBBAND_COUNT_2_HALF; i++) {
			sum = 0;
			for(j = 0; j < SUBBAND_COUNT; j++) {
				sum += *in++ * *m++;	/* 1.3.28 */
			}
			save[offset++] = sum >> 14;	/* 1.1.14 */
			in -= SUBBAND_COUNT;
		}
		offset -= SUBBAND_COUNT_2_HALF;
		in += SUBBAND_COUNT;

		/* 16g̃f[^̏dˍ킹B
		 * |tF[YtB^[oNW\́A炩ߓ]uĂ܂B
		 */
		m = &polyphase_filter[0][0];
		for(i = 0; i < SUBBAND_COUNT_HALF; i++) {
			sum = 0;
			for(j = 0; j < 16 / 2; j++) {
				sum += save[offset] * *m++;	/* 1.3.28 */
				offset += SUBBAND_COUNT_2_HALF + SUBBAND_COUNT_HALF;
				offset &= OFFSET_MASK;
				sum += save[offset] * *m++;	/* 1.3.28 */
				offset += SUBBAND_COUNT_2_HALF - SUBBAND_COUNT_HALF;
				offset &= OFFSET_MASK;
			}
			offset++;
			/* -0x7fff`0x7fffɃXP[OB */
			sum >>= 13;	/* 1.(16).15 */
			if(sum < -0x7fff) sum = -0x7fff;
			if(sum >  0x7fff) sum =  0x7fff;
			*out++ = sum;	/* 1.15 */
		}
		offset -= SUBBAND_COUNT_2_HALF + SUBBAND_COUNT_HALF;
		offset &= OFFSET_MASK;
	}

#undef OFFSET_MASK

#else	/********************************************************************/

	asm volatile ("
	xld.w %r7, 18			; k = SUBBAND_SIZE
mp3_subband_synthesys_L10:
	;
	; ݍ݉ZB(32f[^ԗ̈ɕϊ)
	; ͍s̊s𕄍]ē]u鏈́AIMDCT̒ōsĂ܂B
	;
	xld.w %r11, subband_matrix	; m = subband_matrix
	xld.w %r6, 32			; i = SUBBAND_COUNT_2_HALF
	ld.w %r5, %r14			; save_ptr = &save[offset]
	add %r5, %r15
mp3_subband_synthesys_L20:
	ld.w %alr, %r8			; ݍ
	xld.w %r10, 32			; SUBBAND_COUNT
	mac %r10
	ld.w %r10, %alr			; sum >>= 14
	xsra %r10, 14
	ld.h [%r5]+, %r10		; *save_ptr++ = sum
	xsub %r12, %r12, 64		; in -= SUBBAND_COUNT * sizeof(short)
	;
	xsub %r6, %r6, 1
	xjrne mp3_subband_synthesys_L20
	;
	xadd %r12, %r12, 64		; in += SUBBAND_COUNT * sizeof(short)
	;
	; 16g̃f[^̏dˍ킹B
	; |tF[YtB^[oNW\́A炩ߓ]uĂ܂B
	;
	xld.w %r6, polyphase_filter	; m = polyphse_filter
	xld.w %r11, 16			; i = SUBBAND_COUNT_HALF
mp3_subband_synthesys_L30:
	ld.w %alr, %r8			; sum = 0
	xld.w %r10, 8			; j = 16 / 2
mp3_subband_synthesys_L40:
	ld.w %r5, %r14			; sum += save[offset] * *m++
	add %r5, %r15
	xld.w %r4, 1
	mac %r4
	xadd %r15, %r15, 96		; offset += (SUBBAND_COUNT_2_HALF + SUBBAND_COUNT_HALF) * sizeof(short)
	xand %r15, %r15, 1023		; offset &= OFFSET_MASK (= SUBBAND_COUNT_2_HALF * 16 * sizeof(short) - 1)
	ld.w %r5, %r14			; sum += save[offset] * *m++
	add %r5, %r15
	xld.w %r4, 1
	mac %r4
	xadd %r15, %r15, 32		; offset += (SUBBAND_COUNT_2_HALF - SUBBAND_COUNT_HALF) * sizeof(short)
	xand %r15, %r15, 1023		; offset &= OFFSET_MASK (= SUBBAND_COUNT_2_HALF * 16 * sizeof(short) - 1)
	;
	xsub %r10, %r10, 1
	xjrne mp3_subband_synthesys_L40
	;
	xadd %r15, %r15, 2		; offset += 1 * sizeof(short)
	ld.w %r4, %alr			; sum >>= 13
	xsra %r4, 13
	xcmp %r4, -32767		; if(sum < -0x7fff) sum = -0x7fff
	xjrge mp3_subband_synthesys_L50
	xld.w %r4, -32767
mp3_subband_synthesys_L50:
	xcmp %r4, 32767			; if(sum >  0x7fff) sum =  0x7fff
	xjrle mp3_subband_synthesys_L60
	xld.w %r4, 32767
mp3_subband_synthesys_L60:
	ld.h [%r13]+, %r4		; *out++ = sum
	;
	xsub %r11, %r11, 1
	xjrne mp3_subband_synthesys_L30
	;
	xsub %r15, %r15, 96		; offset -= (SUBBAND_COUNT_2_HALF + SUBBAND_COUNT_HALF) * sizeof(short)
	xand %r15, %r15, 1023		; offset &= OFFSET_MASK (= SUBBAND_COUNT_2_HALF * 16 * sizeof(short) - 1)
	;
	xsub %r7, %r7, 1
	xjrne mp3_subband_synthesys_L10
	;
	ld.w %r10, %r15
	ret
	");

#endif	/********************************************************************/

	return offset;
}

